/* -----------------  flash51n.c ---------------

        Intel Hexfile lesen und anzeigen

   ------------------------------------------------- */


#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#include "minirs232.h"

uint8_t    binbuffer[65536];
int        highestadr, lowestadr;

int        pagebytes;

#define my_usleep(anz)     ({ usleep(anz); fflush(stdout);} )

/* -------------------------- STRDEL ------------------------
     loescht eine Anzahl n Zeichen aus dem String s ab der
     Position pos (inklusive des Zeichens an pos)
     Erstes Zeichen hat Postion (pos) = 1 (und nicht 0) !!!
   ----------------------------------------------------------*/
char *strdel(char *s, int pos, int n)
{
  pos--;
  if (pos> strlen(s)) return s;
  if ((pos+n) > strlen(s))
  {
    n= strlen(s)-pos;
  }
  memmove(s+pos, s+pos+n, strlen(s)-pos-n+1);
  return s;
}

/* -------------------------- STRSPOS ------------------------
     sucht den String srcstr nach dem Substring substr ab
     und liefert die erste Position an der der Substring
     enthalten ist (0 bei Nichtvorhandensein)
   ----------------------------------------------------------*/
int strspos(char *srcstr, char *substr)
// liefert 0 zurueck, wenn Teilstring nicht vorhanden
{
  char *p;
  p= strstr(srcstr, substr);
  if ((p-srcstr+1)> strlen(srcstr)) return 0;
  return (p-srcstr+1);
}

/* -------------------------- STRCPOS ------------------------
     sucht den String srcstr nach dem Zeichen ch ab
     und liefert die erste Position an der das Zeichen
     enthalten ist (0 bei Nichtvorhandensein)
   ----------------------------------------------------------*/
int strcpos(char *srcstr, char ch)
{
  char *p;
  p= strchr(srcstr,ch);
  if ((p-srcstr+1)> strlen(srcstr)) return 0;
  return (p-srcstr+1);
}


/* -------------------------- STRCOPY -----------------------
     kopiert aus dem String src ab der Position pos
     (inklusive) die Anzahl anz Zeichen in den String dest
   ----------------------------------------------------------*/
char *strcopy(char *src, char *dest, int pos, int anz)
{
  int l;

  l= strlen(src);
  if ((pos> l) || (!pos))
  {
    *dest= 0;
    return dest;
  }
  src= src+(pos-1);
  strncpy(dest, src, anz);
  if ( (l-pos+1) < anz) anz= l-pos+1;
  dest[anz]= 0;                         // im Zielstring das Endezeichen setzen
}


/* -------------------- PRINTHEX --------------------

     gibt einen unsigned char als Hex-Wert auf der
     seriellen Schnittstelle aus
   -------------------------------------------------- */

void printhex(uint8_t value)
{
  uint8_t b;

  b= (value >> 4);
  if (b< 10) b += '0'; else b += 'A'-10;
  serial_putchar(b);
  b= value & 0x0f;
  if (b< 10) b += '0'; else b += 'A'-10;
  serial_putchar(b);
}

void printhex16(uint16_t value)
{
  printhex(value >> 8);
  printhex(value & 0x00ff);
}


/* ------------------- SERIAL_GETHEX ----------------

     liest einen Hex-Wert von der seriellen Schnitt-
     stelle ein
   -------------------------------------------------- */
uint8_t serial_gethex(void)
{
  uint8_t ch;
  uint8_t value;

  ch= serial_getchar();
  if (ch > 'F')
  {
    ch= (ch-'a')+10;
  }
  else
  {
    if (ch> '9') ch= (ch-'A')+10; else ch -= '0';
  }
  value= (ch<< 4);
  ch= serial_getchar();

  if (ch > 'F')
  {
    ch= (ch-'a')+10;
  }
  else
  {
    if (ch> '9') ch= (ch-'A')+10; else ch -= '0';
  }
  value |= ch;
  return value;
}

uint16_t serial_gethex16(void)
{
  uint16_t i, i2;

  i= serial_gethex();
  i= i & 0x00ff;
  i2= serial_gethex();
  i2= i2 & 0x00ff;
  i2= i2 | (i << 8);
  return i2;
}

char existfile(char *dname)
{
  FILE    *tdat;

  tdat= fopen(dname,"r");
  if (tdat== 0)
  {
    return 0;
  }
  else
  {
    fclose(tdat);
    return 1;
  }
}

/* ------------------- READHEXFILE ------------------

     liest Intel-Hexdatei ein und speichert die
     Hexwerte in einen Puffer (Speicherspiegel des
     MCU).

     dname  :        zu lesende Datei auf Massenspeicher
     buffer :        Puffer in den binaere Werte ab-
                     gelegt werden
     hexdumpenable : es wird ein Hexdump des
                     Speichers ausgegeb en
   -------------------------------------------------- */
char readhexfile(char* dname, uint8_t *buffer, char hexdumpenable)
{
  FILE     *tdat;
  char     tx[256], cpybuffer[256];
  int      zeil;
  int      b, b2, banz;
  uint8_t  codbyte, cb;
  int      fadress;


  zeil= 0;
  tdat= fopen(dname,"r");
  if (tdat== 0)
  {
    printf("\n\r%s: no such file\n\r",dname);
    return 1;
  }

  b= 0;
  highestadr= 0;
  do
  {
    fgets(tx,255,tdat);

    do                                 // CR entfernen
    {
      b= strcpos(tx,0x0d);
      if (b) strdel(tx,b,1);
    } while (b);
    do                                 // LF entfernen
    {
      b= strcpos(tx,0x0a);
      if (b) strdel(tx,b,1);
    } while (b);

//    printf("\n%s",tx);

    // Hier jede gelesene Zeile der Hexdatei interpretieren und die hoechste
    // und niedrigste Speicheradresse speichern
    if (tx[0]== ':')                   // korrekter Beginn einer Hexzeile ?
    {
      strcopy(tx,cpybuffer,8,2);
      if (!(strcmp(cpybuffer,"00")))   // wenn es normale Datenbytes sind...
      {                                // ... diese auswerten
        zeil++;

        strcopy(tx,cpybuffer,2,2);                // Anzahl Datenbytes in dieser Zeile
        banz= strtol(cpybuffer,0,16);             // hexstring to longint, Basis 16 (Hex)

        strcopy(tx,cpybuffer,4,4);                // Speicheradresse der Datenbytes dieser Zeile
        fadress= strtol(cpybuffer,0,16);

        for (b= 0; b< banz; b++)
        {
          cpybuffer[0]= tx[(b*2)+9];
          cpybuffer[1]= tx[(b*2)+10];
          cpybuffer[2]= 0;
          codbyte= strtol(cpybuffer,0,16);
          binbuffer[fadress]= codbyte;
          fadress++;
          if (fadress> highestadr) highestadr= fadress;
        }
      }
    }

  }while (!feof(tdat));
  fclose(tdat);
  highestadr--;


// Dump -- binbuffer --

  if (hexdumpenable== 1)
  {
    for (b= 0; b< highestadr+1; b++)
    {

      if (!(b % 16) && (b))
      {
        printf("   ");
        for (b2= 16; b2> 0; b2--)
        {
          cb= binbuffer[b-b2];

          if ((cb> 31) && (cb< 127)) printf("%c",cb);
                                else printf(".");

        }
      }
      if (!(b % 16))
      {
        printf("\n%.4X:  ",b);
      }
      printf("%.2X ",binbuffer[b]);

    }
    printf("\n");
  }

  return 0;
}

/* ---------------------- TXPBAR --------------------

     maxwert: der Wert, der 100% entspricht
     value  : aktueller Wert, der angezeigt
              werden soll.
     ch     : Zeichen durch den der Fort-
              schrittsbalken dargestellt
              werden soll.
   -------------------------------------------------- */

void txpbar(uint16_t maxwert, uint16_t value, char ch, uint8_t pbaranzahl, char showtime, float timval)
{
  float r;
  uint8_t b, b2, b3, zanz;

  b2= pbaranzahl;
  for (b= 0; b< b2; b++)
  {
    b3= b % 10;
    if (!b3)
    {
      putchar('|');
    }
    else
    {
      putchar('-');
    }
  }
  if (showtime)
  {
    printf("  %d%% %.2f sec", (value*100)/maxwert, timval);
  }
  printf("\r");
  r= pbaranzahl;
  r= (r / maxwert);
  zanz= r * value;
  for (b= 0; b< zanz; b++)
  {
    putchar(ch);
  }
}

/* --------------------- WRITEFLASH -----------------

     beschreibt den Flashspeicher mit dem Wert in
     binbuffer bis zur Anzahl die in "highestadress"
     hinterlegt ist.
   -------------------------------------------------- */

void writeflash(void)
{
  uint32_t   bf, i;
  uint8_t    dummyb;
  int        error_nr;

  clock_t    start_z, end_z;
  float      diff_z;


  printf("\nwriting flash (%d bytes)\n",highestadr+1);

  serial_putchar('3');                          // Anzahl zu flashende Bytes senden
  printhex16(highestadr);
  i= serial_gethex16();

  serial_putchar('1');                          // Chip loeschen
  dummyb= serial_getchar();
  my_usleep(500000);

  txpbar(highestadr,0,'#',51,1,0);
  serial_putchar('6');                          // Flashvorgang starten

  error_nr= 0;
  start_z= clock();
  for (i= 0; i< highestadr+1; i++)
  {
    bf= binbuffer[i];
    printhex(bf);
    dummyb= serial_gethex();                    // Echo vom Flasher lesen (Ascii)
    if (bf != dummyb) error_nr++;

    end_z= clock(); diff_z= (end_z - start_z); diff_z= diff_z / CLOCKS_PER_SEC;
    printf("\r");
    txpbar(highestadr,i,'#',51,1, diff_z);
    fflush(stdout);
  }
  end_z= clock(); diff_z= (end_z - start_z); diff_z= diff_z / CLOCKS_PER_SEC;
//  printf("  %.2f sec.\n",diff_z);

  printf("\nData transfer errors: %d",error_nr);

}

/* ---------------------- READFLASH -----------------
      liest den Flashspeicher aus und schreibt diese
      in die Datei "filename" im Intel-Hex Format
   -------------------------------------------------- */
void readflash(char *filename)
{
  int        i;
  uint8_t    by, sum;
  int        proglen;
  clock_t    start_z, end_z;
  float      diff_z;

  FILE       *txdat;

  serial_putchar('7');                          // Read chip
  proglen= serial_gethex16();
  if (proglen== 0)
  {
    printf("\nno target device found... \n");
    return;
  }
  printf("\nreading flash (%d bytes)\n",proglen);

  txdat= fopen(filename,"w");
//    printf("\n\r%d\n\r",proglen);
  start_z= clock();
  txpbar(proglen / 32,0,'#',51,1,0);
  for (i= 0; i< proglen; i++)
  {
    end_z= clock(); diff_z= (end_z - start_z); diff_z= diff_z / CLOCKS_PER_SEC;
    printf("\r");
    txpbar(proglen / 32 ,i / 32,'#',51,1, diff_z);
    fflush(stdout);

    if (((i % 32)== 0) && (i != 0))                   // am Ende einer ihex-Zeile "neue Zeile" einfuegen
    {
      sum= ~sum;
      sum++;
      fprintf(txdat,"%.2X\n",sum);                   // Pruefsumme und Zeilenabschluss
    }
    if ((i % 32) == 0)
    {
      sum= 0x20+(i >> 8)+(i & 0xff);

      fprintf(txdat,":20%.2X%.2X00", (i >> 8), (i & 0xff));       // 20 Datenbyte, Adresse, Datetyp 00h (immer 0, weil
                                                                 // kein Speicher groesser als 64k
    }
    by= serial_gethex();
    sum += by;
    fprintf(txdat,"%.2X",by);
  }
  fprintf(txdat,"\n:00000001FF\n");
  fclose(txdat);
}

int main(int argc, char **argv)

{
  uint32_t   b, bf, i;
  uint8_t    dummyb, h;
  char       filename[256];
  char       portname[256];
  char       action[256];
  char       optional[256];
  int        error_nr;

  clock_t    start_z, end_z;
  float      diff_z;


  if (argc >= 3)         // evtl. nur RESET ?
  {
    strcpy(action,argv[1]);
    if (!(strcmp(&action[0],"res")))             // nur Reset ausfuehren !
    {
      strcpy(portname,argv[2]);

      b= existfile(&portname[0]);
      if (!b)
      {
        printf("\n   serialport not found: %s\n\n", portname);
        return 2;
      }

      error_nr= serial_init(&portname[0], 115200);
      if (error_nr)
      {
        printf("\nerror: serialport %s\n\n",portname[0]);
        return 3;
      }

      printf("\nwaiting for sync. adapter...\n");
      if (argc== 5) strcpy(optional,argv[4]); else strcpy(optional,"jjflash");
      if ((strcmp(&optional[0],"nowait")))  my_usleep(1700000);

      serial_putchar('4');                          // Reset chip
      dummyb= serial_getchar();

      printf("\nReset MCU: done ... \n\n");
      serial_close();
      return 0;
    }
  }


  if (argc < 4)
  {
    printf("\n  Syntax:\n  flash51n action port filename [nowait]\n\n");
    printf("  action   : wr   = upload (write) file to mcu\n");
    printf("             rd   = read (download) MCU to file (not yet available)\n");
    printf("             res  = reset microcontroller\n");
    printf("             dump = dump the hexfile\n");
    printf("  port     : adapter ist connected to this port\n");
    printf("  filename : file to flash / read\n\n");
    printf("  nowait   : optional parameter: don\'t delay programstart\n");
    printf("             you can use this option if you have\n");
    printf("             a programer WITHOUT a bootloader\n\n");
    printf("  example  : flash51n wr /dev/ttyUSB0 helloworld.ihx nowait\n\n");
    return 3;
  }

  for (b= 0; b< 0x10000; b++) binbuffer[b]= 0xff;               // Puffer loeschen (nicht geflasht = 0xff)

  if (argc== 5) strcpy(optional,argv[4]); else strcpy(optional,"jjflash");
  strcpy(filename,argv[3]);
  strcpy(portname,argv[2]);
  strcpy(action,argv[1]);


  if (!(strcmp(&action[0],"dump")))
  {
    readhexfile(&filename[0],&binbuffer[0],1);
    return 0;
  }

  if ((strcmp(&action[0],"rd")))                 // wenn Chip nicht gelesen werden soll, Hexdatei von PC lesen
  {
    error_nr= readhexfile(&filename[0],&binbuffer[0], 0);
    if (error_nr) return 1;
  }

  b= existfile(&portname[0]);
  if (!b)
  {
    printf("\n   serialport not found: %s\n\n", portname);
    return 2;
  }

  error_nr= serial_init(&portname[0], 115200);
  if (error_nr)
  {
    printf("\nerror: serialport %s\n\n",portname[0]);
    return 3;
  }

  printf("\nwaiting for sync. adapter...\n\n");

  if ((strcmp(&optional[0],"nowait")))  my_usleep(1700000);

  serial_putchar('4');                          // MCU-Reset
  dummyb= serial_getchar();                     // Antwort Resetkommando lesen
  fflush(stdout);
  my_usleep(100000);
  printf("connected MCU  : ");
  serial_putchar('8');                          // MCU abfragen
  h= serial_gethex();
  switch(h)
  {
    case 0x69 : printf("AT89S51 / S52 (ID broken ?)\n"); pagebytes = 256; break;
    case 0x51 : printf("AT89S51\n"); pagebytes = 256; break;
    case 0x52 : printf("AT89S52\n"); pagebytes = 256; break;
    case 0x23 : printf("AT89S2051\n"); pagebytes = 32; break;
    case 0x43 : printf("AT89S4051\n"); pagebytes = 32; break;
    case 0x73 : printf("AT89S8253\n"); pagebytes = 64; break;
    default:
    {
      printf("not found ! %.2Xh\n",h);
      printf("flash51n finished, thank you...\n\n");
      serial_close();
      return 1;
    }
  }
  printf("pagemode       : %d bytes\n",pagebytes);

  if (!(strcmp(&action[0],"wr")))
  {
    writeflash();
    serial_putchar('4');                          // Reset chip
    dummyb= serial_getchar();
  }

  if (!(strcmp(&action[0],"res")))
  {
    serial_putchar('4');                          // Reset chip
    dummyb= serial_getchar();
  }

  if (!(strcmp(&action[0],"rd")))                 // Read chip
  {
    readflash(&filename[0]);
    serial_putchar('4');
    dummyb= serial_getchar();
  }


  serial_close();
  printf("\n");

  printf("\n");

  return EXIT_SUCCESS;
}

